/*
 * Copyright (C) 2015 Darron Broad
 * All rights reserved.
 * 
 * This file is part of eSPY.
 * 
 * eSPY is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation. 
 * 
 * eSPY is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details. 
 * 
 * You should have received a copy of the GNU General Public License along
 * with eSPY. If not, see http://www.gnu.org/licenses/
 */

/*
 * Various UNIX and Linux header files
 */
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <memory.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>
#include <termios.h>
#include <assert.h>
#include <sysexits.h>
#include <sys/select.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <netdb.h>
#include <net/if.h>

/*
 * Generic buffer
 */
#define BUFLEN (1536)
#define BUFMAX (BUFLEN - 1)

/*
 * I/O time out in seconds
 */
#define TIMEOUT (1)
#define FINAL (3)

/*
 * Baud rate lookup
 */
typedef struct {
	uint32_t baud;
	speed_t speed;
} baudrate_t;

/*
 * LUA interpreter delay
 */
#define SHORTDELAY (150)
#define LONGDELAY (2000)

/*******************************************************************************
 *
 * Open serial device
 *
 ******************************************************************************/
int
openDevice(const char *dev, speed_t speed)
{
	int fd;
	struct termios options;

	fd = open(dev, O_RDWR | O_NOCTTY | O_NONBLOCK);
	if (fd < 0) {
		return fd;
	}
	if (fcntl(fd, F_SETFL, O_NONBLOCK | fcntl(fd, F_GETFL)) < 0) {
		close(fd);
		return -1;
	}
	if (tcgetattr(fd, &options) < 0) {
		close(fd);
		return -1;
	}
#if 1
	 /*
	 * Raw mode 8N2
	 *
	 *  Linux TERMIOS(3)
	 */

	 options.c_iflag &= ~(IGNBRK | BRKINT | PARMRK | ISTRIP | INLCR | IGNCR | ICRNL | IXON);
	 options.c_oflag &= ~(OPOST);
	 options.c_cflag &= ~(CSIZE | PARENB);
	 options.c_cflag |= (CS8 | CSTOPB);
	 options.c_lflag &= ~(ECHO | ECHONL | ICANON | ISIG | IEXTEN);
#else

	/*
	 * Canonical mode 8N2
	 *
	 *  Linux TERMIOS(3)
	 */

	options.c_cflag = CS8 | CSTOPB | CLOCAL | CREAD;
	options.c_iflag = IGNPAR;
	options.c_oflag = 0;
	options.c_lflag = ICANON;
#endif
	options.c_cc[VMIN] = 1;
	options.c_cc[VTIME] = 0;

	if (cfsetispeed(&options, speed) < 0) {
		close(fd);
		return -1;
	}
	if (cfsetospeed(&options, speed) < 0) {
		close(fd);
		return -1;
	}
	if (tcsetattr(fd, TCSANOW, &options) < 0) {
		close(fd);
		return -1;
	}
	if (tcflush(fd, TCIOFLUSH) < 0) {
		close(fd);
		return -1;
	}

	return fd;
}

/*******************************************************************************
 *
 * Return `speed_t' for Given Baud Rate
 *
 ******************************************************************************/
speed_t
serial_speed(uint32_t baudrate)
{
	static baudrate_t rates[] = {
	{0, B0},
	{50, B50},
	{75, B75},
	{110, B110},
	{134, B134},
	{150, B150},
	{200, B200},
	{300, B300},
	{600, B600},
	{1200, B1200},
	{1800, B1800},
	{2400, B2400},
	{4800, B4800},
#ifdef B7200
	{7200, B7200},
#endif
	{9600, B9600},
#ifdef B14400
	{14400, B14400},
#endif
#ifdef B19200
	{19200, B19200},
#else
	{19200, EXTA},
#endif
#ifdef B28800
	{28800, B28800},
#endif
#ifdef B38400
	{38400, B38400},
#else
	{38400, EXTB},
#endif
#ifdef B57600
	{57600, B57600},
#endif
#ifdef B76800
	{76800, B76800},
#endif
#ifdef B115200
	{115200, B115200},
#endif
#ifdef B230400
	{230400, B230400},
#endif
#ifdef B460800
	{460800, B460800},
#endif
#ifdef B500000
	{500000, B500000},
#endif
#ifdef B576000
	{576000, B576000},
#endif
#ifdef B921600
	{921600, B921600},
#endif
#ifdef B1000000
	{1000000, B1000000},
#endif
#ifdef B1152000
	{1152000, B1152000},
#endif
#ifdef B1500000
	{1500000, B1500000},
#endif
#ifdef B2000000
	{2000000, B2000000},
#endif
#ifdef B2500000
	{2500000, B2500000},
#endif
#ifdef B3000000
	{3000000, B3000000},
#endif
#ifdef B3500000
	{3500000, B3500000},
#endif
#ifdef B4000000
	{4000000, B4000000},
#endif
	{UINT32_MAX, B9600},
	};
	int i = 0;

	while (baudrate > rates[i++].baud)
		;

	return rates[--i].speed;
}

/*******************************************************************************
 *
 * Write buflen characters from buffer
 *
 *  return number of bytes or error
 *
 ******************************************************************************/
int
writebuffer(int fd, char *buffer, int buflen, int timeout)
{
	 fd_set fdset;
	 struct timeval timeval;
	 int rc = 0, nb = 0;

	 while (nb < buflen) {
		timeval.tv_sec = timeout;
		timeval.tv_usec = 0;

		FD_ZERO(&fdset);
		FD_SET(fd, &fdset);

		rc = select(fd + 1, NULL, &fdset, NULL, &timeval);
		if (rc < 0) {
		        if (errno == EINTR) {
		                continue;
		        }
		        return -EIO;
		}
		if (rc != 1) {
		        return -ETIMEDOUT;
		}
		if (!FD_ISSET(fd, &fdset)) {
		        return -ETIMEDOUT;
		}
		rc = write(fd, &buffer[nb], buflen - nb);
		if (rc < 0) {
		        if (errno == EINTR || errno == EAGAIN) {
		                continue;
		        }
		        return -EIO;
		}
		if (rc == 0)
		        return -EIO; /* Connection reset */

		nb += rc;
	 }

	 return nb;
}

/*******************************************************************************
 *
 * Read at most buflen characters into buffer
 *
 *  The number of characters read is not guaranteed to be greater than one
 *  regardless of the number requested.
 *
 *  return number of bytes or error
 *
 ******************************************************************************/
int
readchar(int fd, char *buffer, int buflen, int timeout)
{
	 struct timeval timeval;
	 fd_set fdset;
	 int rc = 0;

	 while (1) {
		timeval.tv_sec = timeout;
		timeval.tv_usec = 0;

		FD_ZERO(&fdset);
		FD_SET(fd, &fdset);

		rc = select(fd + 1, &fdset, NULL, NULL, &timeval);
		if (rc < 0) {
		        if (errno == EINTR) {
		                continue;
		        }
		        return -EIO;
		}
		if (rc != 1) {
		        return -ETIMEDOUT;
		}
		if (!FD_ISSET(fd, &fdset)) {
		        return -ETIMEDOUT;
		}
		rc = read(fd, buffer, buflen);
		if (rc < 0) {
		        if (errno == EINTR || errno == EAGAIN) {
		                continue;
		        }
		        return -EIO;
		}
		if (rc == 0)
		        return -EIO; /* Connection reset */

		return rc;
	 }
	 /* Not reached */
}

/*******************************************************************************
 *
 * Read buflen characters into buffer
 *
 *  return number of bytes or error
 *
 ******************************************************************************/
int
readbuffer(int fd, char *buffer, int buflen, int timeout)
{
	 int rc, nb = 0;

	 do {
		rc = readchar(fd, &buffer[nb], buflen - nb, timeout);
		if (rc < 0)
			return rc;
		nb += rc;
	 }
	 while (nb < buflen);

	 return nb;
}

/*******************************************************************************
 *
 * Get quoted command line argument
 *
 ******************************************************************************/
char *
getarg(const char *line, int *len)
{
	*len = strlen(line);
	if (*len == 0)
		return NULL;

	char *arg = calloc(1 + *len, sizeof(char));
	if (arg == NULL) {
		fprintf(stderr, "Fatal error: calloc failed\n");
		exit(EX_OSERR);
	}

	*len = 0;
	while (line[*len] && line[*len] == ' ')
		(*len)++;

	char q;
	if (line[*len] == '\'' || line[*len] == '"')
		q = line[(*len)++];
	else
		q = ' ';

	int j = 0;
	while (line[*len] && line[*len] != q)
		arg[j++] = line[(*len)++];

	if (j == 0) {
		free(arg);
		return NULL;
	}
	arg[j] = '\0';

	return arg;
}

/*******************************************************************************
 *
 * Clean a string for LUA processing
 *
 ******************************************************************************/
char *
LuaClean(char *s, int slen)
{
	int l;
	char *n = s;

	s[slen - 1] = '\0';

	l = strlen((char *)s) - 1;

	while (l >= 0 && s[l] <= ' ')
		s[l--]= '\0';

	while (*n && *n <= ' ')
		n++;

	return n;
}

/*******************************************************************************
 *
 * `strchr' with length limit
 *
 ******************************************************************************/
static char *
strnchr(char *s, char c, int n)
{
        int i = 0;

        while (s[i] && i < n) {
                if (s[i] == c)
                        return &s[i];
                i++;
        }
        return NULL;
}

/*******************************************************************************
 *
 * Read from LUA
 *
 ******************************************************************************/
int
LuaRead(int fd)
{
	static char buffer[BUFLEN];
	static int ix = 0;

	int rc = readchar(fd, &buffer[ix], BUFLEN - ix, TIMEOUT);
	if (rc <= 0)
		return 0;
	ix += rc;

	char *cp;
        while ((cp = strnchr(buffer, '\n', ix)) != NULL) {
	        int i = cp - buffer;

	        buffer[i++] = '\0';
	        ix -= i;

		/*
		 * Prompt filter
		 *
		 * This is problematic of course, but what else can be done?
		 */
		int j = 0;
		while (buffer[j] == 0x3E || buffer[j] == 0x20)
			j++;
		if (buffer[j] != '\r')
		        fprintf(stderr, "%s\n", &buffer[j]);

	        memmove(buffer, &buffer[i], ix);
	}
	return 1;
}

/*******************************************************************************
 *
 * Send a line for LUA processing
 *
 ******************************************************************************/
int
LuaLine(int fd, char *line)
{
	char buffer[BUFLEN];
	int rc, len;

	len = snprintf((char *)buffer, BUFLEN, "%s\r\n", line);
	
	/* WRITE OUT */
	if ((rc = writebuffer(fd, buffer, len, TIMEOUT)) != len) {
		fprintf(stderr, "Write buffer failed [%d] [%s]\n", rc, strerror(errno));
		return -1;
	}

	/* READ IN ECHO */
	if ((rc = readbuffer(fd, buffer, len, TIMEOUT)) != len) {
		fprintf(stderr, "Read buffer failed [%d] [%s]\n", rc, strerror(errno));
		return -1;
	}

	/* LUA LINE DELAY */
	if (strstr((char *)line, "node.restart") == (char *)line) {
		usleep(LONGDELAY * 1000);
	} else {
		usleep(SHORTDELAY * 1000);
	}

	return 0;
}

/*******************************************************************************
 *
 * Run script
 *
 ******************************************************************************/
void
run(FILE *fin, const char *argv)
{
	char *cmd, *dev, *baud;
	int len;
	uint32_t baudrate = 9600;

	int fd;
	char buffer[BUFLEN], *line;
	
	cmd = (char *)strdup(argv);
	if (cmd == NULL) {
		fprintf(stderr, "Fatal error: strdup failed\n");
		exit(EX_OSERR);
	}

	dev = getarg(cmd, &len);
	if (dev == NULL) {
		fprintf(stderr, "Device name missing.\n");
		exit(EX_USAGE);
	}
	cmd += len;

	baud = getarg(cmd, &len);
	if (baud != NULL) {
		baudrate = strtoul(baud, NULL, 0);
		free(baud);
	}
	cmd += len;

	fd = openDevice(dev, serial_speed(baudrate));
	if (fd < 0) {
		fprintf(stderr, "Failed to open I/O device [%s].\n", dev);
		exit(EX_OSERR);
	}

	while (fgets(buffer, BUFLEN, fin) != NULL) {
		line = LuaClean(buffer, BUFLEN);
		if (line[0] == '\0')
			continue;

		LuaLine(fd, line);
		LuaRead(fd);
	}

	int count = FINAL;
	while (count--) {
		if (LuaRead(fd))
			count++;
	}

	close(fd);
}

/*******************************************************************************
 *
 * eSPY - ESP-01 LUA Utility
 *
 ******************************************************************************/
int
main(int argc, char **argv)
{
	FILE *fin;
	char *script, line[BUFLEN];

	if (argc != 3) {
		fprintf(stderr, "Invalid args. [%d]\n", argc);
		exit(EX_USAGE);
	}

	script = argv[2];
	fin = fopen(script, "rb");
	if (fin == NULL) {
		fprintf(stderr, "Failed to open script [%s].\n", script);
		exit(EX_OSERR);
	}

	if (fgets(line, BUFLEN, fin) == NULL) {
		fprintf(stderr, "Failed to read script [%s].\n", script);
		exit(EX_OSERR);
	}
	
	if (line[0] != '#' || line[1] != '!') {
		fprintf(stderr, "Invalid script [%s].\n", script);
		exit(EX_USAGE);
	}

	run(fin, argv[1]);

	fclose(fin);

	exit(EX_OK);
}
